import cvxpy as cp
import numpy as np
import pytest

from dsp import weighted_log_sum_exp
from dsp.cone_transforms import K_repr_x_Gy, LocalToGlob, minimax_to_min
from dsp.problem import MinimizeMaximize, SaddlePointProblem


@pytest.mark.parametrize(
    "x_val,y_val,n", [(1, 1, 2), (2, 2, 2), (0.5, 3, 5), (2, 6, 20), (6, 0.1, 1)]
)
def test_eval_weighted_log_sum_exp(x_val, y_val, n):
    x = cp.Variable(n)

    y = cp.Variable(n, nonneg=True)
    y_value = np.ones(n) * y_val
    x_value = np.ones(n) * x_val

    wlse = weighted_log_sum_exp(x, y)

    lgt = LocalToGlob([x], [y])
    K = wlse.get_K_repr(lgt)

    prob = cp.Problem(
        cp.Minimize(K.f @ y_value + K.t),
        [
            *K.constraints,
            x == x_value,
        ],
    )
    prob.solve(solver=cp.SCS)
    assert prob.status == cp.OPTIMAL
    assert np.isclose(prob.value, np.log(np.exp(x_value) @ y_value), atol=1e-4)


@pytest.mark.parametrize("x_val,y_val,c", [(1, 1, 1), (1, 0.5, 1), (2, 1.23, 2), (3, 2, 3)])
def test_wlse_composition_switch(x_val, y_val, c):
    y = cp.Variable(name="y")
    x1 = cp.Variable(name="x1", nonneg=True)
    x2 = cp.Variable(name="x2", nonneg=True)

    wlse = weighted_log_sum_exp(cp.exp(y), x1 + x2)  # these variables are global scope

    obj = MinimizeMaximize(-wlse)

    x_constraints = [x1 + x2 <= x_val]
    y_constraints = [y >= y_val]

    prob = SaddlePointProblem(obj, x_constraints + y_constraints)
    opt_val = -np.log(x_val * np.exp(np.exp(y_val)))

    prob.x_prob.solve(solver=cp.SCS)
    assert np.isclose(prob.x_prob.value, opt_val)

    prob.solve(solver=cp.SCS)
    assert np.isclose(prob.value, opt_val)


def test_wlse_non_dsp():
    y = cp.Variable(name="y")
    x = cp.Variable(name="x", nonneg=True)

    with pytest.warns(UserWarning, match="Weights are non-positive."):
        assert weighted_log_sum_exp(x, y).is_dsp()
    with pytest.warns(UserWarning, match="Weights are non-positive."):
        assert not weighted_log_sum_exp(cp.log(x), y).is_dsp()
    assert not weighted_log_sum_exp(x, cp.exp(y)).is_dsp()


@pytest.mark.parametrize("x_val,y_val,c", [(1, 1, 1), (1, 0.5, 1), (2.1, 0.5, 2), (1.1, 2, 3)])
def test_wlse_composition(x_val, y_val, c):
    y = cp.Variable(name="y", nonneg=True)
    x1 = cp.Variable(name="x1", nonneg=True)
    x2 = cp.Variable(name="x2", nonneg=True)

    f = cp.square(cp.exp(x1) + x2)
    f_val = (np.exp(x_val) + x_val) ** 2

    # f = cp.square(x1)
    # f_val = (x_val)**2

    # f = cp.exp(x1)
    # f_val = np.exp(x_val)

    wlse = weighted_log_sum_exp(f, y)

    obj = MinimizeMaximize(wlse)

    x_constraints = [x1 >= x_val, x2 >= x_val]
    y_constraints = [y <= y_val]

    prob = SaddlePointProblem(obj, x_constraints + y_constraints)
    assert set(prob.convex_variables()) == {x1, x2}
    assert set(prob.concave_variables()) == {y}
    opt_val = np.log(y_val * np.exp(f_val))

    prob.solve(solver=cp.SCS)
    assert np.isclose(prob.value, opt_val)


def test_eval_weighted_log_sum_exp_affine_sum():
    n = 1
    y_val = 1
    x_val = 1

    x = cp.Variable(n)
    y = cp.Variable(n + 1, nonneg=True)
    y_value = np.ones(n + 1) * y_val
    x_value = np.ones(n) * x_val

    a = np.array([2, 1])
    wlse = weighted_log_sum_exp(x, a @ y + 1)

    ltg = LocalToGlob([x], [y])

    K = wlse.get_K_repr(ltg)

    prob = cp.Problem(
        cp.Minimize(K.f @ y_value + K.t),
        [
            *K.constraints,
            x == x_value,
        ],
    )
    prob.solve(solver=cp.SCS)
    assert prob.status == cp.OPTIMAL
    assert np.isclose(prob.value, np.log(4 * np.exp(1)))

    x_val = 1
    X_constraints = [x >= x_val]
    y_val = 1.1
    Y_constraints = [y <= y_val]
    min_prob = cp.Problem(*minimax_to_min(K, X_constraints, Y_constraints, [y], ltg))
    min_prob.solve(solver=cp.SCS)
    assert min_prob.status == cp.OPTIMAL
    assert np.isclose(min_prob.value, np.log((sum(a) * y_val + 1) * np.exp(x_val)))


@pytest.mark.parametrize("x_val,y_val,c", [(1, 1, 1), (1, 0.5, 1), (1, 1, 1), (0.31, 1.04, 3)])
def test_wlse_multi_var(x_val, y_val, c):
    x1 = cp.Variable(nonneg=True, name="x1")
    x2 = cp.Variable(name="x2")

    y = cp.Variable(3, nonneg=True, name="y")

    a = np.array([2, 1])
    wlse = weighted_log_sum_exp(x1, a @ y[:2] + c)

    obj = MinimizeMaximize(wlse + cp.exp(x2) + cp.log(y[2]))
    x_constraints = [x1 >= x_val, x2 >= x_val]
    y_constraints = [y <= y_val]

    prob = SaddlePointProblem(obj, x_constraints + y_constraints)

    opt_val = np.log((sum(a) * y_val + c) * np.exp(x_val)) + np.exp(x_val) + np.log(y_val)

    prob.x_prob.solve(solver=cp.SCS)
    assert np.isclose(prob.x_prob.value, opt_val, atol=1e-3)

    prob.y_prob.solve(solver=cp.SCS)
    assert np.isclose(prob.y_prob.value, -opt_val)

    prob.solve(solver=cp.SCS)
    assert prob.status == cp.OPTIMAL
    assert np.isclose(prob.value, opt_val, atol=1e-3)
    assert np.allclose(y.value, y_val)


@pytest.mark.parametrize("x_val,y_val,c", [(1, 1, 1), (1, 0.5, 1), (5, 5, 2), (3, 2, 3)])
def test_wlse_switching(x_val, y_val, c):
    y = cp.Variable(name="y")
    x1 = cp.Variable(name="x1", nonneg=True)
    x2 = cp.Variable(name="x2", nonneg=True)

    wlse = weighted_log_sum_exp(y, x1 + x2)  # these variables are global scope

    obj = MinimizeMaximize(-wlse)

    x_constraints = [x1 + x2 <= x_val]
    y_constraints = [y >= y_val]

    prob = SaddlePointProblem(obj, x_constraints + y_constraints)
    opt_val = -np.log(x_val * np.exp(y_val))

    prob.x_prob.solve(solver=cp.SCS)
    assert np.isclose(prob.x_prob.value, opt_val)

    prob.solve(solver=cp.SCS)
    assert np.isclose(prob.value, opt_val)


@pytest.mark.parametrize("x_val,y_val,c", [(1, 1, 1), (1, 0.5, 1), (3.01, 1.4, 2), (3, 2, 3)])
def test_wlse_multi_var_switching(x_val, y_val, c):
    y1 = cp.Variable()
    y2 = cp.Variable()

    x = cp.Variable(3, nonneg=True)

    a = 2
    wlse = weighted_log_sum_exp(a * y1 + a * y2 + c, cp.sum(x[1:]))

    obj = MinimizeMaximize(-wlse + cp.exp(x[0]) + cp.log(y1))
    x_constraints = [x == x_val]
    y_constraints = [y1 == y_val, y2 == y_val]

    prob = SaddlePointProblem(obj, x_constraints + y_constraints)

    opt_val = -np.log(2 * x_val * np.exp(2 * a * y_val + c)) + np.exp(x_val) + np.log(y_val)

    prob.y_prob.solve(solver=cp.SCS)
    assert np.isclose(prob.y_prob.value, -opt_val, atol=1e-2)

    prob.x_prob.solve(solver=cp.SCS)
    assert np.isclose(prob.x_prob.value, opt_val, atol=1e-3)

    prob.solve(solver=cp.SCS)
    assert prob.status == cp.OPTIMAL
    assert np.isclose(prob.value, opt_val, atol=1e-4)


@pytest.mark.parametrize("x_val,y_val,c", [(1, 1, 1)])
def test_neg_wlse(x_val, y_val, c):
    x1 = cp.Variable(nonneg=True, name="x1")
    # x2 = cp.Variable()

    # y = cp.Variable(3, nonneg=True)
    y1 = cp.Variable(nonneg=True, name="y1")

    # a = np.array([2, 1])
    # wlse = WeightedLogSumExp(x1, a@y[:2]+c)
    wlse = weighted_log_sum_exp(y1, x1)

    # obj = MinimizeMaximize(wlse + cp.exp(x2) + cp.log(y[2]))
    obj = MinimizeMaximize(-1 * wlse)
    x_constraints = [x_val >= x1]  # , x2 >= x_val]
    y_constraints = [y_val <= y1]

    prob = SaddlePointProblem(obj, x_constraints + y_constraints)
    prob.solve(solver=cp.SCS)
    assert prob.status == cp.OPTIMAL
    assert np.isclose(prob.value, -np.log(x_val * np.exp(y_val)))


def test_weighted_log_sum_exp_with_switching():
    n = 2
    x = cp.Variable(n, name="x_outer", nonneg=True)
    y = cp.Variable(n, name="y_outer", nonneg=True)

    a = np.ones(n)

    wsle = weighted_log_sum_exp(a @ y + 1, a @ x + 1)

    x_val = 1
    y_val = 3

    X_constraints = [x <= x_val]
    Y_constraints = [y >= y_val]

    ltg = LocalToGlob([x], [y])

    K_switched = wsle.get_K_repr(ltg, switched=True)  # -log(x exp(y))
    min_prob = cp.Problem(*minimax_to_min(K_switched, X_constraints, Y_constraints, [y], ltg))

    min_prob.solve(solver=cp.SCS)
    assert min_prob.status == cp.OPTIMAL
    assert np.isclose(min_prob.value, -np.log((n * x_val + 1) * np.exp(n * y_val + 1)))


@pytest.mark.parametrize("n", [1, 2, 3, 4, 20])
def test_weighted_sum_exp_with_switching(n):
    x = cp.Variable(n, name="x")
    y = cp.Variable(n, name="y", nonneg=True)

    X_constraints = [x == 1]
    Y_constraints = [
        y == 1,
    ]

    # Using x_Gy instead of switched y_Fx
    ltg = LocalToGlob([x], [y])
    K_x_Gy = K_repr_x_Gy(-cp.exp(y), x, ltg)
    min_prob_x_Gy = cp.Problem(*minimax_to_min(K_x_Gy, X_constraints, Y_constraints, [y], ltg))
    min_prob_x_Gy.solve(solver=cp.SCS)
    assert min_prob_x_Gy.status == cp.OPTIMAL
    assert np.isclose(min_prob_x_Gy.value, -n * np.e)


@pytest.mark.parametrize("x_val,y_val", [(1, 2 + np.e + 0.01), (2, 13.01), (1.3, 11.1)])
def test_wlse_comp(x_val, y_val):
    x = cp.Variable(name="x")
    y = cp.Variable(name="y")

    with pytest.warns(UserWarning, match="are non-positive"):
        wlse = weighted_log_sum_exp(cp.square(x), cp.log(cp.log(y)))

    obj = MinimizeMaximize(wlse)

    constraints = [x >= x_val, y <= y_val]

    problem = SaddlePointProblem(obj, constraints)
    opt_val = np.log(np.log(np.log(y_val)) * np.exp(x_val**2))

    problem.solve(solver=cp.SCS)
    assert np.isclose(problem.value, opt_val, atol=1e-4)


def test_wsle_with_external_affine_constraints():
    x = cp.Variable(4, name="x")
    y = cp.Variable(4, name="y", nonneg=True)
    z = cp.Variable(2, name="z")

    wlse = weighted_log_sum_exp(x, y)

    obj = MinimizeMaximize(wlse + z[0] + z[1])

    constraints = [x >= 1, y <= 2, y == np.ones((4, 2)) @ z]

    problem = SaddlePointProblem(obj, constraints)
    opt_val = np.log(8 * np.exp(1)) + 2

    problem.solve(solver=cp.SCS)
    assert np.isclose(problem.value, opt_val, atol=1e-4)


def test_wsle_value():
    n = 2
    x = cp.Variable(n)
    y = cp.Variable(n)

    with pytest.warns(UserWarning, match="Weights are non-positive"):
        wlse = weighted_log_sum_exp(x, y)

    assert wlse.is_incr(0)
    assert wlse.is_incr(1)
    assert not wlse.is_nonneg()
    assert not wlse.is_nonpos()

    assert wlse.value is None

    x.value = np.arange(n)
    y.value = np.arange(n)

    assert wlse.value == np.log(np.sum(y.value * np.exp(x.value)))



from __future__ import annotations

import itertools
from collections import defaultdict
from dataclasses import dataclass, field

import cvxpy as cp
import numpy as np
import scipy.sparse as sp
from cvxpy import SOC
from cvxpy.atoms import reshape
from cvxpy.constraints import ExpCone
from cvxpy.constraints.constraint import Constraint
from cvxpy.constraints.psd import PSD
from cvxpy.expressions.constants import Constant
from cvxpy.expressions.variable import upper_tri_to_full
from cvxpy.problems.objective import Objective
from cvxpy.reductions.dcp2cone.cone_matrix_stuffing import ConeDims


def return_zero() -> float:
    return 0.0


@dataclass
class KRepresentation:
    f: cp.Expression | cp.Variable
    t: cp.Expression | cp.Variable
    constraints: list[Constraint]
    offset: float = 0.0
    y_constraints: list[Constraint] = field(default_factory=list)
    concave_expr: callable = return_zero

    @classmethod
    def sum_of_K_reprs(cls, reprs: list[KRepresentation]) -> KRepresentation:
        assert len(reprs) >= 1

        f = cp.sum([K.f for K in reprs])
        t = cp.sum([K.t for K in reprs])

        all_constraints = [K.constraints for K in reprs]
        constraints = list(itertools.chain.from_iterable(all_constraints))

        offset = np.sum([K.offset for K in reprs])

        def f_concave(x: np.ndarray) -> cp.Expression:
            nones = any(K.concave_expr(x) is None for K in reprs)
            return cp.sum([K.concave_expr(x) for K in reprs]) if not nones else None

        concave_expr = f_concave

        y_constraints = list(itertools.chain.from_iterable([K.y_constraints for K in reprs]))

        return KRepresentation(
            f=f,
            t=t,
            constraints=constraints,
            offset=offset,
            y_constraints=y_constraints,
            concave_expr=concave_expr,
        )

    def scalar_multiply(self, scalar: float) -> KRepresentation:
        assert scalar >= 0
        return KRepresentation(
            f=self.f * scalar,
            t=self.t * scalar,
            constraints=self.constraints,
            offset=self.offset * scalar,
            concave_expr=lambda x: self.concave_expr(x) * scalar
            if self.concave_expr(x) is not None
            else None,
        )

    @classmethod
    def constant_repr(cls, value: float | int) -> KRepresentation:
        if not isinstance(value, (float, int)):
            assert value.size == 1
            assert isinstance(value, np.ndarray)
            value = np.squeeze(value)

        return KRepresentation(
            f=cp.Constant(0),
            t=cp.Constant(0),
            constraints=[],
            offset=float(value),
            concave_expr=lambda x: float(value),
        )


def minimax_to_min(
    K: KRepresentation,
    X_constraints: list[Constraint],
    Y_constraints: list[Constraint],
    y_vars: list[cp.Variable],
    local_to_glob: LocalToGlob,
) -> tuple[Objective, list[Constraint]]:
    # Convex part
    obj = K.t + K.offset

    constraints = [
        *K.constraints,
        *X_constraints,
    ]

    # Concave part
    # this case is only skipped if K.y is zero, i.e., if it's a purely convex problem
    if len(y_vars) > 0:
        all_Y_constraints = Y_constraints + K.y_constraints
        if len(all_Y_constraints) > 0:
            var_id_to_mat, e, cone_dims = get_cone_repr(all_Y_constraints, y_vars)

            n = len(e)
            lamb = cp.Variable(n, name="lamb")
            lamb_const, lamb = add_cone_constraints(lamb, cone_dims, dual=True)

            D = var_id_to_mat["eta"]

            C_shape = (len(e), local_to_glob.y_size)
            C = create_sparse_matrix_from_columns(C_shape, y_vars, local_to_glob, var_id_to_mat)

            if D.shape[1] > 0:
                constraints.append(D.T @ lamb == 0)

            constraints += [C.T @ lamb == K.f, *lamb_const]

            obj += lamb @ e

            unconstrained_vars = [v for v in y_vars if v.id not in var_id_to_mat]
            for v in unconstrained_vars:
                start, end = local_to_glob.var_to_glob[v.id]
                constraints.append(K.f[start:end] == 0)

        else:
            constraints.append(K.f == 0)

    return cp.Minimize(obj), constraints


def scale_psd_dual(cone_dims: ConeDims, lamb: cp.Variable) -> cp.Variable:
    """
    Scale entries of the dual variable lamb corresponding to off-diagonal entries of PSD
    matrices by sqrt(2).
    """

    assert not cone_dims.p3d, "p3d cones not supported"

    if len(cone_dims.psd) > 0:
        exp_offset = 3 * cone_dims.exp
        n = lamb.shape[0]
        scaling_mat_diag = np.ones(n)
        n_psd_entries = sum([d * (d + 1) // 2 for d in cone_dims.psd])
        offset = n - n_psd_entries - exp_offset
        for psd_dim in cone_dims.psd:
            # scale_vec has sqrt(2) on entries corresponding to off-diagonal entries, 1 otherwise
            scale_vec = upper_tri_to_full(psd_dim).A.sum(axis=0) ** 0.5
            compressed_vars = len(scale_vec)
            scaling_mat_diag[offset : offset + compressed_vars] = scale_vec
            offset += compressed_vars
        lamb = np.diag(scaling_mat_diag) @ lamb
    return lamb


def K_repr_x_Gy(
    G: cp.Expression, x: cp.Variable, local_to_glob: LocalToGlob, switched: bool = False
) -> KRepresentation:
    assert G.is_concave()
    assert G.shape == x.shape

    y_vars = G.variables()
    w = cp.Variable(G.size, name="w_bilin")

    constraints = [w <= G]

    (
        var_to_mat_mapping_dual,
        s,
        cone_dims,
    ) = get_cone_repr(constraints, [*y_vars, w])
    Q_bar = var_to_mat_mapping_dual["eta"]
    S_bar = var_to_mat_mapping_dual[w.id]

    lamb = cp.Variable(len(s))
    lamb_constr, lamb = add_cone_constraints(lamb, cone_dims, dual=True)

    t = cp.Variable(name="t_bilin_x_Gy")
    f_size = local_to_glob.y_size if not switched else local_to_glob.x_size
    f = cp.Variable(f_size, name="f_xGy")

    R_bar_shape = (S_bar.shape[0], f_size)
    R_bar = create_sparse_matrix_from_columns(
        R_bar_shape, y_vars, local_to_glob, var_to_mat_mapping_dual
    )

    K_constr = [
        f + R_bar.T @ lamb == 0,
        s @ lamb == t,
        S_bar.T @ lamb == x,
        *lamb_constr,
    ]

    if Q_bar.shape[1] > 0:
        K_constr.append(Q_bar.T @ lamb == 0)

    return KRepresentation(f=f, t=t, constraints=K_constr)


def K_repr_ax(a: cp.Expression) -> KRepresentation:
    assert a.is_convex()
    assert a.size == 1

    f = cp.Variable(name="f_ax")
    t = cp.Variable(name="t_ax")

    constraints = [t >= a, f == 0]

    return KRepresentation(f=f, t=t, constraints=constraints, concave_expr=lambda x: a.value)


class LocalToGlob:
    def __init__(self, x_variables: list[cp.Variable], y_variables: list[cp.Variable]) -> None:
        self.outer_x_vars = x_variables
        self.var_to_glob: dict[int, tuple[int, int]] = {}

        self.x_size = self.add_vars_to_map(x_variables)
        self.y_size = self.add_vars_to_map(y_variables)

    def add_vars_to_map(self, variables: list[cp.Variable]) -> int:
        offset = 0
        for var in variables:
            assert (
                var.ndim <= 1
                or (var.ndim == 2 and min(var.shape) == 1)
                or (var.ndim == 2 and var.shape[0] == var.shape[1])
            )
            # TODO: ensure matrix variables are flattened correctly
            sz = (
                var.size
                if not (var.ndim > 1 and var.is_symmetric())
                else (var.shape[0] * (var.shape[0] + 1) // 2)
            )  # fix for symmetric variables
            if var.attributes["diag"]:
                raise NotImplementedError("Diagonal variables are not supported yet")

            self.var_to_glob[var.id] = (offset, offset + sz)
            offset += sz

        return offset


def K_repr_by(b_neg: cp.Expression, local_to_glob: LocalToGlob) -> KRepresentation:
    assert b_neg.is_concave()
    y_vars = b_neg.variables()
    assert b_neg.size == 1

    b = -b_neg

    t_primal = cp.Variable(name="t_by_primal")

    constraints = [
        t_primal >= b,
    ]

    (
        var_to_mat_mapping,
        s_bar,
        cone_dims,
    ) = get_cone_repr(constraints, [*y_vars, t_primal])

    p_bar = var_to_mat_mapping[t_primal.id]
    Q_bar = var_to_mat_mapping["eta"]

    u = cp.Variable(p_bar.shape[0], name="u_by")
    t = cp.Variable(name="t_by")
    f = cp.Variable(local_to_glob.y_size, name="f_by")

    R_bar_shape = (p_bar.shape[0], local_to_glob.y_size)
    R_bar = create_sparse_matrix_from_columns(
        R_bar_shape, y_vars, local_to_glob, var_to_mat_mapping
    )

    cone_constraints, u = add_cone_constraints(u, cone_dims, dual=True)

    K_constr = [
        f == -R_bar.T @ u,
        t == s_bar @ u,
        p_bar.T @ u + 1 == 0,
        *cone_constraints,
    ]

    if Q_bar.shape[1] > 0:
        K_constr.append(Q_bar.T @ u == 0)

    return KRepresentation(f=f, t=t, constraints=K_constr, concave_expr=lambda x: b_neg)


def K_repr_FxGy(
    Fx: cp.Expression,
    Gy: cp.Expression,
    local_to_glob: LocalToGlob,
    switched: bool = False,
) -> KRepresentation:
    z = cp.Variable(Fx.shape)
    constraints = [z >= Fx]

    K_repr_zGy = K_repr_x_Gy(Gy, z, local_to_glob, switched=switched)

    K_unswitched = KRepresentation(
        f=K_repr_zGy.f, t=K_repr_zGy.t, constraints=constraints + K_repr_zGy.constraints
    )

    if switched:
        return switch_convex_concave(
            K_unswitched.constraints,
            K_repr_zGy.f,
            K_repr_zGy.t,
            Fx.variables(),
            local_to_glob,
        )
    else:
        return K_unswitched


def K_repr_bilin(
    Fx: cp.Expression, Gy: cp.Expression, local_to_glob: LocalToGlob
) -> KRepresentation:
    # Fx = Ax + b, Gy = Cy + d
    # Fx@Gy = Fx.T @ (C y + d)

    C, d = affine_to_canon(Gy, local_to_glob, switched=False)

    if Fx.shape == ():
        Fx = cp.reshape(Fx, (1,))

    return KRepresentation(f=C.T @ Fx, t=Fx.T @ d, constraints=[])


def get_cone_repr(
    const: list[Constraint], variables: list[cp.Variable]
) -> tuple[dict[int, np.ndarray], np.ndarray, ConeDims]:
    assert all(isinstance(v, cp.Variable) for v in variables)

    # TODO: CVXPY does not have a stable API for getting the cone representation that is
    #  solver independent. We use SCS in line with the CVXPY documentation.
    #  Compare https://www.cvxpy.org/tutorial/advanced/index.html#getting-the-standard-form

    aux_prob = cp.Problem(cp.Minimize(0), const)
    solver_opts = {"use_quad_obj": False}

    # get cone representation of c, A, and b for some problem.
    problem_data = aux_prob.get_problem_data(solver_opts=solver_opts, solver=cp.SCS)

    A, const_vec = problem_data[0]["A"].tocsc(), problem_data[0]["b"]

    var_id_to_col = problem_data[0]["param_prob"].var_id_to_col

    unused_mask = np.ones(A.shape[1], dtype=bool)
    var_to_mat_mapping = {}

    for v in variables:
        if v.id not in var_id_to_col:
            continue
        start_ind = var_id_to_col[v.id]
        sz = (
            v.size
            if not (v.ndim > 1 and v.is_symmetric())
            else (v.shape[0] * (v.shape[0] + 1) // 2)
        )  # fix for symmetric variables
        end_ind = start_ind + sz
        original_cols = np.arange(start_ind, end_ind)

        var_to_mat_mapping[v.id] = A[:, original_cols]
        unused_mask[original_cols] = 0

    var_to_mat_mapping["eta"] = A[:, unused_mask]

    cone_dims = problem_data[0]["dims"]

    return var_to_mat_mapping, const_vec, cone_dims


def add_cone_constraints(
    s: cp.Expression, cone_dims: ConeDims, dual: bool
) -> tuple[list[Constraint], cp.Variable]:
    assert len(s.shape) == 1 or s.shape[1] == 1, "s must be a vector"
    s = cp.reshape(s, (s.shape[0],))

    s_const = []

    offset = 0
    if cone_dims.zero > 0:
        if not dual:
            s_const.append(s[: cone_dims.zero] == 0)
        offset += cone_dims.zero
    if cone_dims.nonneg > 0:
        s_const.append(s[offset : offset + cone_dims.nonneg] >= 0)
        offset += cone_dims.nonneg
    if len(cone_dims.soc) > 0:
        t_inds_dict = defaultdict(list)
        X_inds_dict = defaultdict(list)
        for soc_dim in cone_dims.soc:
            t_inds_dict[soc_dim].append(offset)
            X_inds_dict[soc_dim].extend(range(offset + 1, offset + soc_dim))
            offset += soc_dim

        for soc_dim, t_inds in t_inds_dict.items():
            X_slice = s[X_inds_dict[soc_dim]]
            X = cp.reshape(X_slice, (soc_dim - 1, len(t_inds)), order="F")
            s_const.append(SOC(t=s[t_inds], X=X, axis=0))

    if len(cone_dims.psd) > 0:
        for psd_dim in cone_dims.psd:
            m = psd_dim * (psd_dim + 1) // 2
            z = s[offset : offset + m]

            fill_coeff = Constant(upper_tri_to_full(psd_dim))
            flat_mat = fill_coeff @ z
            full_mat = reshape(flat_mat, (psd_dim, psd_dim))
            s_const.append(PSD(full_mat))
            offset += m

    if cone_dims.exp > 0:
        end = offset + 3 * cone_dims.exp
        if dual:
            tau = s[offset + 2 : end : 3]  # z (in cvxpy) -> t -> tau
            sigma = s[offset + 1 : end : 3]  # y (in cvxpy) -> s -> sigma
            rho = -s[offset:end:3]  # x (in cvxpy) -> r -> -rho
            s_const.extend([tau >= 0, rho >= 0, sigma >= cp.rel_entr(rho, tau) - rho])
        else:
            x = s[offset:end:3]
            y = s[offset + 1 : end : 3]
            z = s[offset + 2 : end : 3]
            s_const.append(ExpCone(x, y, z))

        offset += 3 * cone_dims.exp

    if len(cone_dims.p3d) > 0:
        raise NotImplementedError

    assert offset == s.size

    lamb = scale_psd_dual(cone_dims, s)
    return s_const, lamb


def affine_to_canon(
    expr: cp.Expression, local_to_glob: LocalToGlob, switched: bool
) -> tuple[np.ndarray, np.ndarray]:
    variables = expr.variables()
    aux = cp.Variable(expr.shape)

    (
        var_to_mat_mapping,
        c,
        cone_dims,
    ) = get_cone_repr([aux == expr], [*variables, aux])

    # get the equality constraints
    rows_needed = cone_dims.zero
    assert rows_needed == expr.size

    cols = local_to_glob.y_size if not switched else local_to_glob.x_size

    rows_present = len(c)
    B_shape = (rows_present, cols)
    B = create_sparse_matrix_from_columns(B_shape, variables, local_to_glob, var_to_mat_mapping)
    B = B[:rows_needed]

    aux_columns = var_to_mat_mapping[aux.id][:rows_needed]
    assert (np.abs(aux_columns) - sp.eye(aux.size)).nnz == 0
    assert (np.sign(aux_columns.diagonal()) == np.ones(rows_needed)).all() or (
        np.sign(aux_columns.diagonal()) == -np.ones(rows_needed)
    ).all()

    sgn = -np.sign(aux_columns.diagonal())[0]
    B = B * sgn

    c = c[:rows_needed]

    return B, c


def split_K_repr_affine(
    expr: cp.Expression, convex_vars: list[cp.Variable], concave_vars: list[cp.Variable]
) -> tuple[cp.Expression, cp.Expression, cp.Constant]:
    """
    Split an affine expression into a convex and concave part plus offset, i.e.
    A@x+b <=> C@convex_vars + D@concave_vars + b
    """
    assert expr.is_affine()
    aux = cp.Variable(expr.shape)
    var_to_mat_mapping, b, _ = get_cone_repr([aux == expr], [*convex_vars, *concave_vars])
    C = cp.Constant(0)
    D = cp.Constant(0)
    for v in convex_vars:
        C += -var_to_mat_mapping.get(v.id, 0) @ cp.vec(v)
    for v in concave_vars:
        D += -var_to_mat_mapping.get(v.id, 0) @ cp.vec(v)

    return C, D, cp.Constant(b)


def create_sparse_matrix_from_columns(
    shape: tuple[int, int],
    variables: list[cp.Variable],
    local_to_glob: LocalToGlob,
    var_to_mat_mapping: dict[int, sp.csc_matrix],
) -> sp.csc_matrix:
    cols = []
    for v in variables:
        start, end = local_to_glob.var_to_glob[v.id]
        cols.append((start, end, v.id))

    cols = sorted(cols, key=lambda x: x[0])

    mats_to_stack = []
    current_col = 0
    for start, end, v_id in cols:
        if start != current_col:
            mats_to_stack.append(sp.csc_matrix((shape[0], start - current_col), dtype=float))
        mat = var_to_mat_mapping.get(v_id, sp.csc_matrix((shape[0], end - start), dtype=float))
        mats_to_stack.append(mat)
        current_col = end
    if current_col < shape[1]:
        mats_to_stack.append(sp.csc_matrix((shape[0], shape[1] - current_col)))

    stacked_mat = sp.hstack(mats_to_stack, format="csc")
    assert stacked_mat.shape == shape
    return stacked_mat


def switch_convex_concave(
    constraints: list[Constraint],
    f: cp.Variable,
    t: cp.Variable,
    x_vars: list[cp.Variable],
    local_to_glob: LocalToGlob,
    precomp: cp.Variable | None = None,
) -> KRepresentation:
    """
    Return switched global k_repr from k_repr constraints and x/y expressions
    """

    var_list = [f, t] + x_vars

    var_to_mat_mapping, s, cone_dims = get_cone_repr(constraints, var_list)

    u_bar = cp.Variable(len(s))
    u_bar_const, u_bar = add_cone_constraints(u_bar, cone_dims, dual=True)

    P = var_to_mat_mapping[f.id]
    p = var_to_mat_mapping[t.id].toarray().flatten()
    Q = var_to_mat_mapping["eta"]

    R_shape = (len(s), local_to_glob.y_size)
    R = create_sparse_matrix_from_columns(R_shape, x_vars, local_to_glob, var_to_mat_mapping)

    f_bar = cp.Variable(local_to_glob.y_size, name="f_bar")
    t_bar = cp.Variable(name="t_bar")

    constraints = [
        f_bar == -R.T @ u_bar,
        t_bar == s @ u_bar,
        # P.T @ u_bar + x_bar == 0,
        *u_bar_const,
        p @ u_bar + 1 == 0,
        Q.T @ u_bar == 0,
    ]

    if precomp is not None:
        constraints += [P.T @ u_bar + precomp == 0]
    else:
        for v in local_to_glob.outer_x_vars:
            start, end = local_to_glob.var_to_glob[v.id]
            # P.T @ u_bar + x_bar == 0,

            if v.ndim > 1 and v.is_symmetric():
                n = v.shape[0]
                inds = np.triu_indices(n, k=0)  # includes diagonal
                v = v[inds]

            constraints += [(P.T @ u_bar)[start:end] + cp.vec(v) == 0]

    return KRepresentation(f=f_bar, t=t_bar, constraints=constraints)



from __future__ import annotations

from typing import Any

import cvxpy as cp
from cvxpy.constraints.constraint import Constraint

from dsp.parser import DSPError
from dsp.problem import semi_infinite_epigraph
from dsp.saddle_extremum import saddle_max, saddle_min


def saddle_max_canon(expr: saddle_max, _args: Any) -> tuple[cp.Expression, list[Constraint]]:
    if not expr.is_dsp():
        raise DSPError("The objective function must be a DSP expression.")
    mode = "sup"
    obj, aux_constraints = semi_infinite_epigraph(
        expr.f, expr.concave_variables(), expr.constraints, mode
    )
    return obj, aux_constraints


def saddle_min_canon(expr: saddle_min, _args: Any) -> tuple[cp.Expression, list[Constraint]]:
    if not expr.is_dsp():
        raise DSPError("The objective function must be a DSP expression.")
    mode = "inf"
    obj, aux_constraints = semi_infinite_epigraph(
        expr.f, expr.convex_variables(), expr.constraints, mode
    )
    return -obj, aux_constraints  # -obj because we want a hypograph variable



import cvxpy as cp
import pytest

from dsp import is_dsp, saddle_max, saddle_min, weighted_log_sum_exp
from dsp.local import LocalVariable, LocalVariableError
from dsp.parser import DSPError
from dsp.problem import MinimizeMaximize, SaddlePointProblem


def test_is_dsp():
    x = cp.Variable(2, name="x", nonneg=True)
    y = cp.Variable(2, name="y", nonneg=True)
    z = cp.Variable(name="z")
    f = weighted_log_sum_exp(x, x)

    obj = MinimizeMaximize(f)
    saddle_problem = SaddlePointProblem(obj, [cp.sum(x) == 1])

    assert (
        not saddle_problem.is_dsp()
    )  # AssertionError: Cannot add variables to both convex and concave set.

    obj = MinimizeMaximize(weighted_log_sum_exp(x, y) + z)
    saddle_problem = SaddlePointProblem(obj, [cp.sum(x) == 1, cp.sum(y) == 1])

    assert (
        not saddle_problem.is_dsp()
    )  # AssertionError: Cannot resolve curvature of variables ['z']. Specify curvature of these variables as SaddleProblem(obj, constraints, minimization_vars, maximization_vars).

    obj = MinimizeMaximize(weighted_log_sum_exp(x, y))
    saddle_problem = SaddlePointProblem(
        obj, [cp.sum(x) == 1, cp.sum(y) == 1], minimization_vars=[x, y]
    )

    assert (
        not saddle_problem.is_dsp()
    )  # AssertionError: Cannot add variables to both convex and concave set.

    with pytest.raises(DSPError, match="Cannot add variables to both convex and concave set."):
        saddle_problem.solve()


def test_concave_saddle_max():
    y_dummy = LocalVariable(2, name="y_dummy", nonneg=True)
    x = cp.Variable(name="x", nonneg=True)
    f = -cp.sqrt(cp.sum(y_dummy)) + cp.exp(x)
    F = saddle_max(f, [cp.sum(y_dummy) == 1])
    assert not F.is_dsp()


def test_saddle_double_dummy():
    x = cp.Variable(name="x", nonneg=True)
    y_local = LocalVariable(name="y", nonneg=True)

    y = cp.Variable(name="y", nonneg=True)
    x_local = LocalVariable(name="x", nonneg=True)

    f = cp.sqrt(y_local) + cp.exp(x)
    g = cp.sqrt(y) + cp.exp(x_local)

    F_1 = saddle_max(f, [cp.sum(y_local) == 1])

    with pytest.raises(LocalVariableError):
        saddle_max(f, [cp.sum(y_local) == 1])

    G_1 = saddle_min(g, [cp.sum(x_local) == 1])
    with pytest.raises(LocalVariableError):
        saddle_min(g, [cp.sum(x_local) == 1])

    F_1.is_dsp()
    G_1.is_dsp()

    prob = cp.Problem(cp.Minimize(F_1), [F_1 <= 0])
    assert is_dsp(prob)
    prob_2 = cp.Problem(cp.Maximize(G_1), [G_1 >= 0])
    assert is_dsp(prob_2)


def test_problem_is_dcp():
    x = cp.Variable(2, name="x", nonneg=True)
    y_local = LocalVariable(2, name="y", nonneg=True)

    f = weighted_log_sum_exp(x, y_local)
    sup_y_f = saddle_max(f, [cp.sum(y_local) == 1])

    prob = cp.Problem(cp.Minimize(sup_y_f), [sup_y_f <= 0])

    assert is_dsp(prob)
    assert prob.is_dsp()

    y_local = LocalVariable(2, name="y", nonneg=True)
    f = weighted_log_sum_exp(x, y_local)
    f += cp.exp(y_local[1])
    sup_y_f = saddle_max(f, [cp.sum(y_local) == 1])
    prob = cp.Problem(cp.Minimize(sup_y_f), [sup_y_f <= 0])

    assert not is_dsp(prob)
    assert not prob.is_dsp()


def test_expression_is_dcp():
    x = cp.Variable(2, name="x", nonneg=True)
    y_local = LocalVariable(2, name="y", nonneg=True)

    f = weighted_log_sum_exp(x, y_local)
    sup_y_f = saddle_max(f, [cp.sum(y_local) == 1])
    F = sup_y_f + cp.exp(x[1])

    assert is_dsp(sup_y_f)
    assert sup_y_f.is_dsp()

    assert F.is_dsp()

    y_local = LocalVariable(2, name="y", nonneg=True)
    f = weighted_log_sum_exp(x, y_local)
    f += cp.exp(y_local[1])
    sup_y_f = saddle_max(f, [cp.sum(y_local) == 1])
    F = sup_y_f + cp.exp(x[1])

    assert not is_dsp(sup_y_f)
    assert not sup_y_f.is_dsp()
    assert not F.is_dsp()


def test_saddle_fun_is_dsp():
    x = cp.Variable(2, name="x", nonneg=True)
    y = LocalVariable(2, name="y", nonneg=True)
    f = weighted_log_sum_exp(x, y)
    z = cp.Variable(name="z")

    f1 = f + z
    assert f1.is_dsp()
    assert saddle_max(f1, [cp.sum(y) == 1]).is_dsp()

    assert (f - z).is_dsp()
    assert not (f * z).is_dsp()
    assert not (f + cp.exp(z) + cp.sqrt(z)).is_dsp()


def test_saddle_extremum_affine_is_dsp():
    x = cp.Variable(2, name="x", nonneg=True)
    y = LocalVariable(2, name="y", nonneg=True)
    wlse = weighted_log_sum_exp(x, y)
    z = cp.Variable(name="z")

    f = wlse + z
    assert f.is_dsp()

    F = saddle_max(f, [cp.sum(y) == 1])
    assert F.is_dsp()

    x = LocalVariable(2, name="x", nonneg=True)
    y = cp.Variable(2, name="y", nonneg=True)
    wlse = weighted_log_sum_exp(x, y)
    z = cp.Variable(name="z")

    f = wlse + z
    assert f.is_dsp()

    F = saddle_min(f, [cp.sum(x) == 1])
    assert F.is_dsp()

    assert not F.is_nonpos()
    assert not F.is_nonneg()

    # Arguments are the variables of the expression, no monotonicities implemented
    assert not F.is_incr(0)
    assert not F.is_decr(0)


def test_saddle_extremum_local_affine():
    x = cp.Variable(2, name="x", nonneg=True)
    y = LocalVariable(2, name="y_local", nonneg=True)
    z = LocalVariable(name="z_local")
    f = weighted_log_sum_exp(x, y) + z

    assert f.is_dsp()

    F = saddle_max(f, [cp.sum(y) == 1])
    assert (
        F.is_dsp()
    )  # z is a local variable appearing in the objective, and not listed as a concave variable

    x = LocalVariable(2, name="x_local", nonneg=True)
    y = cp.Variable(2, name="y", nonneg=True)
    z = LocalVariable(name="z_local")
    f = weighted_log_sum_exp(x, y) + z

    G = saddle_min(f, [cp.sum(x) == 1])
    assert (
        G.is_dsp()
    )  # z is a local variable appearing in the objective, and not listed as a convex variable


def test_saddle_extremum_missing_local():
    x = cp.Variable(2, name="x", nonneg=True)
    y_local = LocalVariable(2, name="y_local1", nonneg=True)
    z = cp.Variable(name="z")
    f = weighted_log_sum_exp(x, y_local) + z

    assert f.is_dsp()

    F = saddle_max(f, [cp.sum(y_local) == 1])
    assert F.is_dsp()

    # max
    x = cp.Variable(2, name="x", nonneg=True)
    y_local = LocalVariable(2, name="y_local2", nonneg=True)
    f = weighted_log_sum_exp(x, y_local) + cp.log(z)
    F = saddle_max(f, [cp.sum(y_local) == 1])

    assert f.is_dsp()
    assert not F.is_dsp()

    x = cp.Variable(2, name="x", nonneg=True)
    y_local = LocalVariable(2, name="y_local2", nonneg=True)
    f = weighted_log_sum_exp(x, y_local) + cp.exp(z)
    F = saddle_max(f, [cp.sum(y_local) == 1])

    assert f.is_dsp()
    assert F.is_dsp()


def test_bad_other():
    x_local = LocalVariable(2, name="x_local", nonneg=True)
    y = cp.Variable(2, name="y_new", nonneg=True)
    z = cp.Variable(name="z")
    f = weighted_log_sum_exp(x_local, y) + cp.log(z)

    F = saddle_min(f, [cp.sum(x_local) == 1])
    assert f.is_dsp()
    assert F.is_dsp()

    x_local = LocalVariable(2, name="x_local", nonneg=True)
    f = weighted_log_sum_exp(x_local, y) + cp.exp(z)
    F = saddle_min(f, [cp.sum(x_local) == 1])
    assert f.is_dsp()
    assert not F.is_dsp()


def test_bad_curvatures():
    x = cp.Variable(2, name="x", nonneg=True)
    y_local = LocalVariable(2, name="y_local", nonneg=True)
    y1_local = LocalVariable(name="y1_local", nonneg=True)
    z = cp.Variable(name="z")
    f = weighted_log_sum_exp(x, y_local) + z

    F = saddle_max(f, [cp.sum(y_local) == y1_local, y1_local == 1])
    assert f.is_dsp()
    assert F.is_dsp()


def test_affine_parts():
    x = cp.Variable(2, name="x", nonneg=True)
    y_local = LocalVariable(2, name="y_local", nonneg=True)
    z = cp.Variable(name="z")
    f = weighted_log_sum_exp(x, y_local) + z
    assert f.is_dsp()

    F = saddle_max(f, [cp.sum(y_local) == 1])
    assert F.is_dsp()
    assert set(F.convex_variables()) == {x, z}

    y_local = LocalVariable(2, name="y_local", nonneg=True)
    z_local = LocalVariable(name="z_local")
    f = weighted_log_sum_exp(x, y_local) + z_local
    assert f.is_dsp()

    F = saddle_max(f, [cp.sum(y_local) == 1])
    assert F.is_dsp()

    y_local = LocalVariable(2, name="y_local", nonneg=True)
    z_local = LocalVariable(name="z_local")
    f = weighted_log_sum_exp(x, y_local) + z_local
    F = saddle_max(f, [cp.sum(y_local) == 1])
    assert F.is_dsp()


def test_saddle_extremum_non_dcp_constraint():
    x = cp.Variable(2, name="x", nonneg=True)
    y_local = LocalVariable(2, name="y_local", nonneg=True)

    f = weighted_log_sum_exp(x, y_local)
    F = saddle_max(f, [cp.sum(cp.abs(y_local)) == 1])
    assert not F.is_dsp()

    with pytest.raises(cp.DCPError):
        cp.Problem(cp.Minimize(F), [x == 0]).solve()


def test_unknown_type():
    assert not is_dsp("Not a cvxpy expression")



from __future__ import annotations

from itertools import chain
from typing import Iterable

import cvxpy as cp
import numpy as np
from cvxpy.constraints.constraint import Constraint
from cvxpy.problems.objective import Objective
from cvxpy.utilities import Canonical

from dsp.cone_transforms import (
    LocalToGlob,
    add_cone_constraints,
    get_cone_repr,
    minimax_to_min,
)
from dsp.parser import DSPError, Parser, initialize_parser
from dsp.saddle_extremum import SaddleExtremum


class MinimizeMaximize(Canonical):
    def __init__(self, expr: cp.Expression) -> None:
        self._validate_arguments(expr)
        self.args = [cp.Expression.cast_to_const(expr)]

    @staticmethod
    def _validate_arguments(expr: cp.Expression | float | int) -> None:
        if isinstance(expr, cp.Expression):
            assert expr.size == 1
        else:
            assert isinstance(expr, (float, int)), f"Cannot parse {expr=}"

    def is_dsp(self) -> bool:
        return self.expr.is_dsp()

    @property
    def value(self) -> float:
        return self.expr.value


class SaddlePointProblem(cp.Problem):
    def __init__(
        self,
        minmax_objective: MinimizeMaximize,
        constraints: list[Constraint] | None = None,
        minimization_vars: Iterable[cp.Variable] | None = None,
        maximization_vars: Iterable[cp.Variable] | None = None,
    ) -> None:
        self._validate_arguments(minmax_objective)
        self.obj_expr = minmax_objective.expr

        # Optional inputs
        self._minimization_vars = set(minimization_vars) if minimization_vars is not None else set()
        self._maximization_vars = set(maximization_vars) if maximization_vars is not None else set()
        self._constraints_ = list(constraints) if constraints is not None else []  # copy

        self._x_prob = None
        self._y_prob = None

        self._value: float | None = None
        self._status: str | None = None
        super().__init__(cp.Minimize(self.obj_expr), constraints)

    @property
    def x_prob(self) -> cp.Problem:
        if self._x_prob is None:
            constraints_x, single_obj_x = self.dualized_problem(
                self.obj_expr, self._constraints_, self._minimization_vars, self._maximization_vars
            )
            self._x_prob = cp.Problem(single_obj_x, constraints_x)
        return self._x_prob

    @property
    def y_prob(self) -> cp.Problem:
        if self._y_prob is None:
            # note the variables are switched, and the problem value will be negated
            constraints_y, single_obj_y = self.dualized_problem(
                -self.obj_expr, self._constraints_, self._maximization_vars, self._minimization_vars
            )
            self._y_prob = cp.Problem(single_obj_y, constraints_y)
        return self._y_prob

    def dualized_problem(
        self,
        obj_expr: cp.Expression,
        constraints: list[Constraint],
        minimization_vars: Iterable[cp.Variable],
        maximization_vars: Iterable[cp.Variable],
    ) -> tuple[list[Constraint], Objective]:
        parser = initialize_parser(obj_expr, minimization_vars, maximization_vars, constraints)

        local_to_glob_y = LocalToGlob(parser.convex_vars, parser.concave_vars)

        K_repr = parser.parse_expr_repr(obj_expr, switched=False, local_to_glob=local_to_glob_y)

        single_obj, constraints = minimax_to_min(
            K_repr,
            parser.x_constraints,
            parser.y_constraints,
            list(parser.concave_vars),
            local_to_glob_y,
        )

        return constraints, single_obj

    @staticmethod
    def _validate_arguments(minmax_objective: MinimizeMaximize) -> None:
        assert isinstance(minmax_objective, MinimizeMaximize)

    def solve(self, eps: float = 1e-3, *args, **kwargs: dict) -> float:  # noqa
        """
        Solves the saddle point problem.
        """
        self.x_prob.solve(*args, **kwargs)
        assert self.x_prob.status in {cp.OPTIMAL, cp.OPTIMAL_INACCURATE}, self.x_prob.status

        self.y_prob.solve(*args, **kwargs)
        assert self.y_prob.status in {cp.OPTIMAL, cp.OPTIMAL_INACCURATE}, self.y_prob.status

        diff = self.x_prob.value + self.y_prob.value  # y_prob.value is negated
        assert np.isclose(self.x_prob.value, -self.y_prob.value, atol=eps) and np.isclose(
            -self.y_prob.value, self.x_prob.value, atol=eps
        ), f"Difference between x and y problem is {diff}, (should be 0)."

        self._status = (
            cp.OPTIMAL_INACCURATE
            if cp.OPTIMAL_INACCURATE in {self.x_prob.status, self.y_prob.status}
            else cp.OPTIMAL
        )
        self._value = self.x_prob.value
        return self._value

    @property
    def status(self) -> str | None:
        return self._status

    @property
    def value(self) -> float | None:
        """
        Returns the value of the objective function at the solution, and None if the problem has not
        been successfully solved.
        """
        return self._value

    def is_dsp(self) -> bool:
        # try to form x_prob and catch the exception
        try:
            self.x_prob  # noqa
            return True
        except DSPError:
            return False

    def convex_variables(self) -> list[cp.Variable]:
        parser = initialize_parser(
            self.obj_expr, self._minimization_vars, self._maximization_vars, self._constraints_
        )
        return sorted(parser.convex_vars, key=lambda x: x.id)

    def concave_variables(self) -> list[cp.Variable]:
        parser = initialize_parser(
            self.obj_expr, self._minimization_vars, self._maximization_vars, self._constraints_
        )
        return sorted(parser.concave_vars, key=lambda x: x.id)

    def affine_variables(self) -> list[cp.Variable]:
        parser = initialize_parser(
            self.obj_expr, self._minimization_vars, self._maximization_vars, self._constraints_
        )
        return sorted(parser.affine_vars, key=lambda x: x.id)


def semi_infinite_epigraph(
    expr: cp.Expression, variables: list[cp.Variable], constraints: list[Constraint], mode: str
) -> tuple[cp.Expression, list[Constraint]]:
    assert mode in ["inf", "sup"], "Not a valid semi-infinite mode."

    minimization_vars = variables if mode == "inf" else []
    maximization_vars = variables if mode == "sup" else []

    aux_prob = SaddlePointProblem(
        MinimizeMaximize(expr), constraints, minimization_vars, maximization_vars
    )
    # aux_prob.solve()  # TODO: should we solve here to verify saddle point property?
    # assert aux_prob.status == cp.OPTIMAL
    prob = aux_prob.x_prob if mode == "sup" else aux_prob.y_prob
    obj = prob.objective.expr
    aux_constraints = prob.constraints  # aux_constraints may not be canonicalized

    variables = list(set(chain.from_iterable([c.variables() for c in aux_constraints])))
    var_id_map = {v.id: v for v in variables}
    var_to_mat_mapping, const_vec, cone_dims = get_cone_repr(aux_constraints, variables)

    # A @ [all variables]
    expr = 0
    aux_size = var_to_mat_mapping["eta"].shape[1]
    if aux_size > 0:  # used extra variables
        eta = cp.Variable(aux_size)
        expr += var_to_mat_mapping["eta"] @ eta
    var_to_mat_mapping.pop("eta")

    for v_id, A_ in var_to_mat_mapping.items():
        v = var_id_map[v_id]
        if v.ndim > 1 and v.is_symmetric():
            # create v_vec, the symmetric part of v
            assert v.shape[0] == v.shape[1]
            inds = np.triu_indices(v.shape[0], k=0)  # includes diagonal
            expr += A_ @ v[inds]
        else:
            expr += A_ @ cp.vec(v)

    z = const_vec - expr  # Ax + b in K

    cone_constraints, z = add_cone_constraints(z, cone_dims, dual=False)

    return obj, cone_constraints


def get_problem_SE_atoms(problem: cp.Problem) -> list[SaddleExtremum]:
    SE_atoms = []
    SE_atoms += get_SE_atoms(problem.objective)
    for constraint in problem.constraints:
        SE_atoms += get_SE_atoms(constraint)
    return SE_atoms


def get_SE_atoms(expr: Canonical) -> list[SaddleExtremum]:
    if isinstance(expr, SaddleExtremum):
        return [expr]
    elif not expr.args:
        return []
    else:
        return list(chain.from_iterable([get_SE_atoms(arg) for arg in expr.args]))


def validate_saddle_extremum(
    SE_atom: SaddleExtremum, problem_constraints: list[Constraint]
) -> None:
    aux_prob = SaddlePointProblem(
        MinimizeMaximize(SE_atom.f),
        constraints=problem_constraints + SE_atom.constraints,
        minimization_vars=SE_atom.convex_variables(),
        maximization_vars=SE_atom.concave_variables(),
    )
    aux_prob.solve()
    assert aux_prob.status in {cp.OPTIMAL, cp.OPTIMAL_INACCURATE}


def validate_all_saddle_extrema(problem: cp.Problem) -> None:
    SE_atoms = get_problem_SE_atoms(problem)
    for SE_atom in SE_atoms:
        validate_saddle_extremum(SE_atom, problem.constraints)


def is_dsp_expr(obj: cp.Expression) -> bool:
    if obj.is_dcp():
        return True
    try:
        parser = Parser(set(), set())
        parser.parse_expr_variables(obj, switched=False)
        return True
    except DSPError:
        return False


def is_dsp(obj: cp.Problem | SaddlePointProblem | cp.Expression) -> bool:
    if isinstance(obj, SaddlePointProblem):
        return obj.is_dsp()
    elif isinstance(obj, cp.Problem):
        all_SE_atoms = get_problem_SE_atoms(obj)
        return obj.is_dcp() and all(atom.is_dsp() for atom in all_SE_atoms)
    elif isinstance(obj, cp.Expression):
        return is_dsp_expr(obj)
    else:
        return False
